import struct
from dataclasses import dataclass, field
from typing import List

import factom_core.blockchains.mainnet
import factom_core.primitives as primitives
from factom_core.blockchains import Blockchain
from factom_core.blocks import (
    DirectoryBlock,
    AdminBlock,
    FactoidBlock,
    EntryCreditBlock,
    EntryBlock,
)
from factom_core.block_elements import Entry
from factom_core.messages import Message


@dataclass
class DirectoryBlockState(Message):
    """
    A message to communicate a Directory Block State
    """

    TYPE = 20

    timestamp: bytes
    directory_block: DirectoryBlock
    admin_block: AdminBlock
    factoid_block: FactoidBlock
    entry_credit_block: EntryCreditBlock
    entry_blocks: List[EntryBlock]
    entries: List[Entry]
    signatures: primitives.FullSignatureList

    # Not marshalled
    is_in_database: bool = field(init=False, default=False)

    def __post_init__(self):
        # TODO: type/value assertions
        self.is_p2p = True
        super().__post_init__()

    def marshal(self) -> bytes:
        """
        Marshal the message into the following representation:
        - first byte is the message type (always 20)
        - next bytes are the marshalled Directory Block
        - next bytes are the marshalled Admin Block
        - next bytes are the marshalled Factoid Block
        - next bytes are the marshalled Entry Credit Block
        - next 4 bytes are the number of Entry Blocks following (NOTE: always 0 now)
        - next bytes are the aforementioned marshalled Entry Blocks
        - next 4 bytes are the number of Entries following (NOTE: always 0 now)
        - next bytes are all of the Entries, marshalled as:
            - 4 bytes size of the Entry
            - marshalled Entry itself
        - next bytes are the number of signatures
        - next bytes are all of the signatures marshalled as:
            - 64 bytes signature
            - 32 byte public key

        :return: byte representation of the message
        """
        buf = bytearray()
        buf.append(self.TYPE)
        buf.extend(self.timestamp)
        buf.extend(self.directory_block.marshal())
        buf.extend(self.admin_block.marshal())
        buf.extend(self.factoid_block.marshal())
        buf.extend(self.entry_credit_block.marshal())
        buf.extend(struct.pack(">I", len(self.entry_blocks)))
        for entry_block in self.entry_blocks:
            buf.extend(entry_block.marshal())
        buf.extend(struct.pack(">I", len(self.entries)))
        for entry in self.entries:
            entry_data = entry.marshal()
            buf.extend(struct.pack(">I", len(entry_data)))
            buf.extend(entry_data)
        buf.extend(self.signatures.marshal())
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        msg_type, data = raw[0], raw[1:]
        if msg_type != cls.TYPE:
            raise ValueError("Invalid message type ({})".format(msg_type))

        timestamp, data = data[:6], data[6:]
        directory_block, data = DirectoryBlock.unmarshal_with_remainder(data)
        admin_block, data = AdminBlock.unmarshal_with_remainder(data)
        factoid_block, data = FactoidBlock.unmarshal_with_remainder(data)
        entry_credit_block, data = EntryCreditBlock.unmarshal_with_remainder(data)

        entry_block_count, data = struct.unpack(">I", data[:4])[0], data[4:]
        entry_blocks = []
        for i in range(entry_block_count):
            entry_block, data = EntryBlock.unmarshal_with_remainder(data)
            entry_blocks.append(entry_block)

        entry_count, data = struct.unpack(">I", data[:4])[0], data[4:]
        entries = []
        for i in range(entry_count):
            entry_size, data = struct.unpack(">I", data[:4])[0], data[4:]
            entry, data = Entry.unmarshal(data[:entry_size]), data[entry_size:]
            entries.append(entry)

        signatures = primitives.FullSignatureList.unmarshal(data)

        return DirectoryBlockState(
            timestamp=timestamp,
            directory_block=directory_block,
            admin_block=admin_block,
            factoid_block=factoid_block,
            entry_credit_block=entry_credit_block,
            entry_blocks=entry_blocks,
            entries=entries,
            signatures=signatures,
        )

    def to_dict(self):
        return {
            "timestamp": self.timestamp.hex(),
            "directory_block": self.directory_block.to_dict(),
            "admin_block": self.admin_block.to_dict(),
            "factoid_block": self.factoid_block.to_dict(),
            "entry_credit_block": self.entry_credit_block.to_dict(),
            "entry_blocks": [v.to_dict() for v in self.entry_blocks],
            "entries": [v.to_dict() for v in self.entries],
            "signatures": [sig.to_dict() for sig in self.signatures],
        }

    def is_sane(self, state: Blockchain):
        if None in {
            self.directory_block,
            self.admin_block,
            self.factoid_block,
            self.entry_credit_block,
        }:
            return False

        if self.is_in_database:
            return True

        height = self.directory_block.header.height
        if height == 0:
            return True  # Always accept genesis blocks
        # TODO: if height < state.height_at_boot: return False
        # TODO: if height < state.highest_saved_block: return False

        if state.network_id != self.directory_block.header.network_id:
            return False

        if type(state) is factom_core.blockchains.mainnet.MainnetBlockchain:
            checkpoints = factom_core.blockchains.mainnet.constants.CHECKPOINTS
            keymr = checkpoints.get(height)
            if keymr is not None:
                if self.directory_block.keymr.hex() != keymr:
                    return False
        return True

    def is_valid(self):
        # TODO: dbstate.validate_signatures()

        # Hash checks for all blocks in Directory Block Body
        body = self.directory_block.body
        if body.admin_block_lookup_hash != self.admin_block.lookup_hash:
            return False
        if body.factoid_block_keymr != self.factoid_block.keymr:
            return False
        if body.entry_credit_block_header_hash != self.entry_credit_block.header_hash:
            return False

        # Make dict of stuff claimed to be in directory block
        entry_block_claims = set()
        entry_claims = set()
        for entry_block in body.entry_blocks:
            entry_block_claims.add(entry_block.keymr)

        # Check claims against actual included entry blocks
        for entry_block in self.entry_blocks:
            if entry_block.keymr not in entry_block_claims:
                return False
            # Add claims from the included entry blocks
            claims = {entry_hash for hashes in entry_block.body.entry_hashes.values() for entry_hash in hashes}
            entry_claims.update(claims)

        # Check claims of entry blocks against actual included entries
        for entry in self.entries:
            if entry.entry_hash not in entry_claims:
                return False

        return True

    def leader_execute(self, state: Blockchain):
        self.follower_execute(state)

    def follower_execute(self, state: Blockchain):
        pass


@dataclass
class DirectoryBlockStateRequest(Message):
    """
    A request for missing DBState objects.
    """

    TYPE = 21

    timestamp: bytes
    block_height_start: int
    block_height_end: int

    def __post_init__(self):
        # TODO: type/value assertions
        self.is_p2p = True
        super().__post_init__()

    def marshal(self) -> bytes:
        """
        Marshal the message into the following representation:
        - first byte is the message type (always 21)
        - next 6 bytes are the timestamp
        - next 4 bytes are the block height start
        - next 4 bytes are the block height end

        :return: byte representation of the message
        """
        buf = bytearray()
        buf.append(self.TYPE)
        buf.extend(self.timestamp)
        buf.extend(struct.pack(">I", self.block_height_start))
        buf.extend(struct.pack(">I", self.block_height_end))
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        msg_type, data = raw[0], raw[1:]
        if msg_type != cls.TYPE:
            raise ValueError("Invalid message type ({})".format(msg_type))

        timestamp, data = data[:6], data[6:]
        block_height_start, data = struct.unpack(">I", data[:4])[0], data[4:]
        block_height_end, data = struct.unpack(">I", data[:4])[0], data[4:]

        return DirectoryBlockStateRequest(
            timestamp=timestamp, block_height_start=block_height_start, block_height_end=block_height_end,
        )

    def to_dict(self):
        return {
            "timestamp": self.timestamp.hex(),
            "block_height_start": self.block_height_start,
            "block_height_end": self.block_height_end,
        }

    def leader_execute(self, state: Blockchain):
        self.follower_execute(state)

    def follower_execute(self, state: Blockchain):
        pass


@dataclass()
class BlockRequest(Message):
    """
    An unimplemented message. I assume for requesting single blocks of any type.
    """

    TYPE = 14

    timestamp: bytes

    def __post_init__(self):
        # TODO: type/value assertions
        super().__post_init__()

    def marshal(self) -> bytes:
        """
        Marshal the message into the following representation:
        - first byte is the message type (always 14)
        - next 6 bytes are a timestamp

        :return: byte representation of the message
        """
        buf = bytearray()
        buf.append(self.TYPE)
        buf.extend(self.timestamp)
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        msg_type, data = raw[0], raw[1:]
        if msg_type != cls.TYPE:
            raise ValueError("Invalid message type ({})".format(msg_type))
        timestamp, data = data[:6], data[6:]
        assert len(data) == 0, "Extra bytes remaining!"
        return BlockRequest(timestamp=timestamp)

    def to_dict(self):
        return {"timestamp": self.timestamp.hex()}



from typing import Any, List

import factom_core.blocks as blocks
from factom_core.db import FactomdLevelDB

from .pending_block import PendingBlock


class BaseBlockchain:
    """The base class for all Blockchain objects"""

    network_id: bytes = None
    vms: List[Any] = None

    data_path: str = None
    db: FactomdLevelDB = None
    current_block: PendingBlock = None

    def __init__(self, data_path: str = None) -> None:
        if not isinstance(self.network_id, bytes) or len(self.network_id) != 4:
            raise ValueError("The Blockchain class must be instantiated with a `network_id` bytes object of length 4")
        # if not isinstance(self.vms, list) or len(self.vms) == 0:
        #     raise ValueError(
        #         "The Blockchain class must be instantiated with a `vms` list of length > 1"
        #     )
        self.data_path = data_path
        self.db = FactomdLevelDB(path=data_path, create_if_missing=True)

    def load_genesis_block(self) -> blocks.DirectoryBlock:
        raise NotImplementedError("Blockchain classes must implement this method")

    def vm_for_hash(self, h: bytes) -> int:
        raise NotImplementedError("Blockchain classes must implement this method")

    def seal_minute(self) -> None:
        raise NotImplementedError("Blockchain classes must implement this method")

    def rotate_vms(self) -> None:
        raise NotImplementedError("Blockchain classes must implement this method")

    def seal_block(self) -> None:
        raise NotImplementedError("Blockchain classes must implement this method")


class Blockchain(BaseBlockchain):
    """
    A Blockchain is a combination of VM classes. Each VM is associated
    with a range of chains. The Blockchain class acts as a wrapper around these other
    VM classes, delegating operations to the appropriate VM depending on the
    current block / minute number.
    """

    def __init__(self, data_path: str = None):
        super().__init__(data_path)

    def load_genesis_block(self) -> blocks.DirectoryBlock:
        pass

    def vm_for_hash(self, h: bytes) -> int:
        """
        Compute the VM index responsible for hash h

        Taken from: factomd/state/processList.go/VMindexFor(hash []byte)
        """
        if len(self.vms) == 0:
            return 0
        v = 0
        for b in h:
            v += b
        return v % len(self.vms)

    def seal_minute(self) -> None:
        """Finalize the current block minute"""
        self.rotate_vms()

        if self.current_block.current_minute == 10:
            self.seal_block()
        else:
            self.current_block.current_minute += 1

    def rotate_vms(self) -> None:
        """Rotate the responsibilities of the VM set (if necessary)"""
        # TODO: see processList.go/MakgeMap for formula per block height
        if len(self.vms) == 1:
            return
        self.vms = self.vms[1:] + self.vms[:1]

    def seal_block(self):
        """
        Bundles all added transactions, entries, and other elements into a set of finalized
        blocks with headers.
        """
        block = self.current_block
        entry_blocks: List[blocks.EntryBlock] = []
        for chain_id, block_body in block.entry_blocks.items():
            prev = self.db.get_entry_block_head(chain_id)
            header = block_body.construct_header(
                chain_id=chain_id,
                prev_keymr=prev.keymr if prev is not None else bytes(32),
                prev_full_hash=prev.full_hash if prev is not None else bytes(32),
                sequence=prev.header.sequence + 1 if prev is not None else 0,
                height=block.height,
            )
            entry_blocks.append(blocks.EntryBlock(header, block_body))

        prev = self.db.get_entry_credit_block(height=block.height - 1)
        header = block.entry_credit_block.construct_header(
            prev_header_hash=prev.header_hash, prev_full_hash=prev.full_hash, height=block.height,
        )
        entry_credit_block = blocks.EntryCreditBlock(header, block.entry_credit_block)

        prev = self.db.get_factoid_block(height=block.height - 1)
        header = block.factoid_block.construct_header(
            prev_keymr=block.previous.body.factoid_block_keymr,
            prev_ledger_keymr=prev.ledger_keymr,
            ec_exchange_rate=1000,  # TODO
            height=block.height,
        )
        factoid_block = blocks.FactoidBlock(header, block.factoid_block)

        prev = self.db.get_admin_block(height=block.height - 1)
        header = block.admin_block.construct_header(back_reference_hash=prev.back_reference_hash, height=block.height)
        admin_block = blocks.AdminBlock(header, block.admin_block)

        # Compile all the above blocks and the previous directory block, into a new one
        directory_block_body = blocks.DirectoryBlockBody(
            admin_block_lookup_hash=admin_block.lookup_hash,
            entry_credit_block_header_hash=entry_credit_block.header_hash,
            factoid_block_keymr=factoid_block.keymr,
            entry_blocks=[
                {"chain_id": entry_block.header.chain_id, "keymr": entry_block.keymr} for entry_block in entry_blocks
            ],
        )
        header = directory_block_body.construct_header(
            network_id=self.network_id,
            prev_keymr=block.previous.keymr,
            prev_full_hash=block.previous.full_hash,
            timestamp=block.timestamp,
            height=block.height,
        )
        directory_block = blocks.DirectoryBlock(header, directory_block_body)

        # Persist the blocks as new chain heads
        self.db.put_directory_block_head(directory_block)
        self.db.put_admin_block_head(admin_block)
        self.db.put_entry_credit_block_head(entry_credit_block)
        self.db.put_factoid_block_head(factoid_block)
        for entry_block in entry_blocks:
            self.db.put_entry_block_head(entry_block)



import hashlib
from dataclasses import dataclass

import factom_core.primitives as primitives
from factom_core.utils import varint


@dataclass
class FactoidTransaction:

    timestamp: int
    inputs: list
    outputs: list
    ec_purchases: list
    rcds: primitives.FullSignatureList

    def __post_init__(self):
        # TODO: assert they're all here
        pass

    def is_coinbase(self):
        # TODO: Coinbase outputs be zero too right? Just matters that everything else is definitely zero
        return len(self.inputs) == 0 and len(self.ec_purchases) == 0 and len(self.rcds) == 0

    @property
    def hash(self):
        return hashlib.sha256(self.marshal()).digest()

    @property
    def tx_id(self):
        return hashlib.sha256(self.marshal_for_signature()).digest()

    def marshal(self):
        """Marshals the FactoidTransaction according to the byte-level representation shown at
        https://github.com/FactomProject/FactomDocs/blob/master/factomDataStructureDetails.md#factoid-transaction
        """
        buf = bytearray()
        buf.extend(self.marshal_for_signature())
        for rcd in self.rcds:
            buf.append(0x01)
            buf.extend(rcd.marshal())
        return bytes(buf)

    def marshal_for_signature(self):
        """Marshals the transaction's header and partial body, in order to be signed by a key"""
        buf = bytearray()
        buf.append(0x02)
        buf.extend(self.timestamp.to_bytes(6, "big", signed=False))
        buf.append(len(self.inputs))
        buf.append(len(self.outputs))
        buf.append(len(self.ec_purchases))
        for i in self.inputs:
            value = i.get("value")
            fct_address = i.get("fct_address")
            buf.extend(varint.encode(value) + fct_address)
        for o in self.outputs:
            value = o.get("value")
            fct_address = o.get("fct_address")
            buf.extend(varint.encode(value) + fct_address)
        for purchase in self.ec_purchases:
            value = purchase.get("value")
            ec_public_key = purchase.get("ec_public_key")
            buf.extend(varint.encode(value) + ec_public_key)
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        """Returns a new FactoidTransaction object, unmarshalling given bytes according to:
        https://github.com/FactomProject/FactomDocs/blob/master/factomDataStructureDetails.md#factoid-transaction
        """
        obj, data = FactoidTransaction.unmarshal_with_remainder(raw)
        assert len(data) == 0, "Extra bytes remaining!"
        return obj

    @classmethod
    def unmarshal_with_remainder(cls, raw: bytes):
        """Returns a new FactoidTransaction object, unmarshalling given bytes according to:
        https://github.com/FactomProject/FactomDocs/blob/master/factomDataStructureDetails.md#factoid-transaction

        Because FactoidTransaction is variable length, it is useful to pass a "stream" of bytes to be unmarshalled.
        This way, we don't have to know the size in advance. Just return the remainder bytes for the caller to use
        elsewhere.
        """
        data = raw[1:]  # skip single byte version, probably just 0x02 anyways
        timestamp, data = int.from_bytes(data[:6], "big", signed=False), data[6:]
        input_count, data = ord(data[:1]), data[1:]
        output_count, data = ord(data[:1]), data[1:]
        ec_purchase_count, data = ord(data[:1]), data[1:]

        inputs = []
        for i in range(input_count):
            value, data = varint.decode(data)
            fct_address, data = data[:32], data[32:]
            inputs.append({"value": value, "fct_address": fct_address})

        outputs = []
        for i in range(output_count):
            value, data = varint.decode(data)
            fct_address, data = data[:32], data[32:]
            outputs.append({"value": value, "fct_address": fct_address})

        ec_purchases = []
        for i in range(ec_purchase_count):
            value, data = varint.decode(data)
            ec_public_key, data = data[:32], data[32:]
            ec_purchases.append({"value": value, "ec_public_key": ec_public_key})

        rcds = primitives.FullSignatureList()
        for i in range(input_count):
            data = data[1:]  # skip 1 byte version number, always 0x01 for now
            signature, data = primitives.FullSignature.unmarshal(data[:96]), data[96:]
            rcds.append(signature)

        return (
            FactoidTransaction(
                timestamp=timestamp, inputs=inputs, outputs=outputs, ec_purchases=ec_purchases, rcds=rcds,
            ),
            data,
        )

    def to_dict(self):
        return {
            "tx_id": self.tx_id.hex(),
            "timestamp": self.timestamp,
            "inputs": [{"value": r.get("value"), "fct_address": r.get("fct_address").hex()} for r in self.inputs],
            "outputs": [{"value": r.get("value"), "fct_address": r.get("fct_address").hex()} for r in self.outputs],
            "ec_purchases": [
                {"value": r.get("value"), "ec_public_key": r.get("ec_public_key").hex()} for r in self.ec_purchases
            ],
            "rcds": [r.to_dict() for r in self.rcds],
        }

    def __str__(self):
        # TODO: convert timestamp to readable
        return "{}(timestamp={}, inputs={}, outputs={}, ec_purchases={})".format(
            self.__class__.__name__, self.timestamp, len(self.inputs), len(self.outputs), len(self.ec_purchases),
        )



from dataclasses import dataclass
from hashlib import sha256
from .entry import Entry


@dataclass
class Chain:

    chain_id: bytes
    first_entry: Entry

    def __post_init__(self):
        # TODO: do we need these safety checks?
        assert self.chain_id == self._calculate_chain_id(), "chain_id does not match external_ids"
        assert isinstance(self.first_entry, Entry), "first_entry must be of type models.Entry"

    def _calculate_chain_id(self):
        """Returns the chain id in bytes. The algorithm used is shown at:
        https://github.com/FactomProject/FactomDocs/blob/master/factomDataStructureDetails.md#chainid
        """
        # Chain ID = SHA256(SHA256(ExtIDs[0]) + SHA256(ExtIDs[0]) + ... + SHA256(ExtIDs[N]))
        chain_id = sha256()
        for x in self.first_entry.external_ids:
            chain_id.update(sha256(x).digest())
        return chain_id.digest()

    def to_dict(self):
        return {"chain_id": self.chain_id, "first_entry": self.first_entry}

    def __str__(self):
        return "{}(chain_id={})".format(self.__class__.__name__, self.chain_id.hex())



import hashlib
from factom_core.blockchains import Blockchain


class CustomBlockchain(Blockchain):
    def __init__(self, network_name: str, data_path: str) -> None:
        self.network_id = hashlib.sha256(network_name.encode()).digest()[:4]
        super().__init__(data_path)



from .base import Message, unmarshal_message
from .ack import Ack
from .block_syncing import DirectoryBlockState, DirectoryBlockStateRequest, BlockRequest
from .entry_syncing import MissingDataRequest, MissingDataResponse
from .heartbeat import Heartbeat
from .missing_message import MissingMessageRequest, MissingMessageResponse
from .seals import DirectoryBlockSignature, EndOfMinute
from .server import AddServer, ChangeServerKey, RemoveServer
from .transactions import FactoidTransaction, ChainCommit, EntryCommit, EntryReveal



import struct
from dataclasses import dataclass

from factom_core.blockchains import Blockchain
from factom_core.messages import Message


@dataclass
class MissingMessageRequest(Message):
    """
    A request for a missing message in a node's process list
    """

    TYPE = 16

    timestamp: bytes
    asking: bytes
    vm_index: int
    height: int
    system_height: int
    process_list_heights: list

    def __post_init__(self):
        # TODO: type/value assertions
        self.is_p2p = True
        super().__post_init__()

    def marshal(self) -> bytes:
        """
        Marshal the message into the following representation:
        - first byte is the message type (always 16)
        - next 6 bytes are the timestamp
        - next 32 bytes are the asker hash (?)
        - next byte is the VM index
        - next 4 bytes are the directory block height
        - next 4 bytes are the system height
        - next 4 bytes are the number of missing message process list heights
        - next X bytes are a series of 4 byte process list heights

        :return: byte representation of the message
        """
        buf = bytearray()
        buf.append(self.TYPE)
        buf.extend(self.timestamp)
        buf.extend(self.asking)
        buf.append(self.vm_index)
        buf.extend(struct.pack(">I", self.height))
        buf.extend(struct.pack(">I", self.system_height))
        buf.extend(struct.pack(">I", len(self.process_list_heights)))
        for h in self.process_list_heights:
            buf.extend(struct.pack(">I", h))
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        msg_type, data = raw[0], raw[1:]
        if msg_type != cls.TYPE:
            raise ValueError("Invalid message type ({})".format(msg_type))

        timestamp, data = data[:6], data[6:]
        asking, data = data[:32], data[32:]
        vm_index, data = data[0], data[1:]
        height, data = struct.unpack(">I", data[:4])[0], data[4:]
        system_height, data = struct.unpack(">I", data[:4])[0], data[4:]
        process_list_heights_count, data = struct.unpack(">I", data[:4])[0], data[4:]
        process_list_heights = []
        for _ in range(process_list_heights_count):
            h, data = struct.unpack(">I", data[:4])[0], data[4:]
            process_list_heights.append(h)

        return MissingMessageRequest(
            timestamp=timestamp,
            asking=asking,
            vm_index=vm_index,
            height=height,
            system_height=system_height,
            process_list_heights=process_list_heights,
        )

    def to_dict(self) -> dict:
        return {
            "timestamp": self.timestamp.hex(),
            "asking": self.asking.hex(),
            "vm_index": self.vm_index,
            "height": self.height,
            "system_height": self.system_height,
            "process_list_heights": [v for v in self.process_list_heights],
        }

    def leader_execute(self, state: Blockchain):
        self.follower_execute(state)

    def follower_execute(self, state: Blockchain):
        pass


@dataclass
class MissingMessageResponse(Message):
    """
    A response to a MissingMessageRequest, containing the requested message
    """

    TYPE = 19

    timestamp: bytes

    def __post_init__(self):
        # TODO: type/value assertions
        self.is_p2p = True
        super().__post_init__()

    def marshal(self) -> bytes:
        """
        Marshal the message into the following representation:
        - first byte is the message type (always 19)
        - next 6 bytes are the timestamp

        :return: byte representation of the message
        """
        buf = bytearray()
        buf.append(self.TYPE)
        buf.extend(self.timestamp)
        # TODO: figure out the rest of this MissingMessageResponse
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        msg_type, data = raw[0], raw[1:]
        if msg_type != cls.TYPE:
            raise ValueError("Invalid message type ({})".format(msg_type))

        timestamp, data = data[:6], data[6:]
        b, data = data[0], data[1:]
        # TODO: figure out the rest of this MissingMessageResponse

        return MissingMessageResponse(timestamp=timestamp)

    def leader_execute(self, state: Blockchain):
        self.follower_execute(state)

    def follower_execute(self, state: Blockchain):
        pass



import datetime


def get_milli_time(delta: int = 0):
    t = int((datetime.datetime.now().timestamp() + delta) * 1000)
    return int.to_bytes(t, 6, "big", signed=False)


def datetime_from_milli_time(timestamp: bytes):
    t = int.from_bytes(timestamp, "big", signed=False) / 1000
    return datetime.datetime.fromtimestamp(t)



import bottle
import json
import sys
import factom_core.messages
import factom_core.db
from enum import Enum


bottle.BaseRequest.MEMFILE_MAX = 1024 * 1024
app = bottle.default_app()

rest_path = "/rest/v1"


class RestPaths(Enum):
    DIRECTORY_BLOCK = f"{rest_path}/dblocks"
    ADMIN_BLOCK = f"{rest_path}/ablocks"
    FACTOID_BLOCK = f"{rest_path}/fblocks"
    ENTRY_CREDIT_BLOCK = f"{rest_path}/ecblocks"
    ENTRY_BLOCK = f"{rest_path}/eblocks"
    ENTRY = f"{rest_path}/entries"


hex_regex = "[0-9A-Fa-f]{64}"


@bottle.hook("before_request")
def strip_path():
    """Strip trailing '/' on all requests. '/foo' and /foo/' are two unique endpoints in bottle"""
    bottle.request.environ["PATH_INFO"] = bottle.request.environ["PATH_INFO"].rstrip(
        "/"
    )


@bottle.get("/health")
def health_check():
    return {"data": "Healthy!"}


@bottle.get(f"{RestPaths.DIRECTORY_BLOCK.value}/<keymr:re:{hex_regex}>")
def get_directory_block(keymr: str):
    db = factom_core.db.FactomdLevelDB(create_if_missing=True)
    block = db.get_directory_block(keymr=bytes.fromhex(keymr))
    db.close()
    if block is None:
        bottle.abort(404)
    return block.to_dict()


@app.get(f"{RestPaths.DIRECTORY_BLOCK.value}/<height:int>")
def get_directory_block_by_height(height: int):
    db = factom_core.db.FactomdLevelDB(create_if_missing=True)
    block = db.get_directory_block(height=height)
    db.close()
    if block is None:
        bottle.abort(404)
    return block.to_dict()


@app.get(f"{RestPaths.DIRECTORY_BLOCK.value}/head")
def get_directory_block_head():
    db = factom_core.db.FactomdLevelDB(create_if_missing=True)
    block = db.get_directory_block_head()
    db.close()
    if block is None:
        bottle.abort(404)
    return block.to_dict()


@bottle.get(f"{RestPaths.ADMIN_BLOCK.value}/<lookup_hash:re:{hex_regex}>")
def get_admin_block(lookup_hash: str):
    db = factom_core.db.FactomdLevelDB(create_if_missing=True)
    block = db.get_admin_block(lookup_hash=bytes.fromhex(lookup_hash))
    db.close()
    if block is None:
        bottle.abort(404)
    return block.to_dict()


@app.get(f"{RestPaths.ADMIN_BLOCK.value}/<height:int>")
def get_admin_block_by_height(height: int):
    db = factom_core.db.FactomdLevelDB(create_if_missing=True)
    block = db.get_admin_block(height=height)
    db.close()
    if block is None:
        bottle.abort(404)
    return block.to_dict()


@bottle.get(f"{RestPaths.ADMIN_BLOCK.value}/head")
def get_admin_block_head():
    db = factom_core.db.FactomdLevelDB(create_if_missing=True)
    block = db.get_admin_block_head()
    db.close()
    if block is None:
        bottle.abort(404)
    return block.to_dict()


@bottle.get(f"{RestPaths.FACTOID_BLOCK.value}/<keymr:re:{hex_regex}>")
def get_factoid_block(keymr: str):
    db = factom_core.db.FactomdLevelDB(create_if_missing=True)
    block = db.get_factoid_block(keymr=bytes.fromhex(keymr))
    db.close()
    if block is None:
        bottle.abort(404)
    return block.to_dict()


@bottle.get(f"{RestPaths.FACTOID_BLOCK.value}/<height:int>")
def get_factoid_block_by_height(height: int):
    db = factom_core.db.FactomdLevelDB(create_if_missing=True)
    block = db.get_factoid_block(height=height)
    db.close()
    if block is None:
        bottle.abort(404)
    return block.to_dict()


@bottle.get(f"{RestPaths.FACTOID_BLOCK.value}/head")
def get_factoid_block_head():
    db = factom_core.db.FactomdLevelDB(create_if_missing=True)
    block = db.get_factoid_block_head()
    db.close()
    if block is None:
        bottle.abort(404)
    return block.to_dict()


@bottle.get(f"{RestPaths.ENTRY_CREDIT_BLOCK.value}/<header_hash:re:{hex_regex}>")
def get_entry_credit_block(header_hash: str):
    db = factom_core.db.FactomdLevelDB(create_if_missing=True)
    block = db.get_entry_credit_block(keymr=bytes.fromhex(header_hash))
    db.close()
    if block is None:
        bottle.abort(404)
    return block.to_dict()


@bottle.get(f"{RestPaths.ENTRY_CREDIT_BLOCK.value}/<height:int>")
def get_entry_credit_block_by_height(height: int):
    db = factom_core.db.FactomdLevelDB(create_if_missing=True)
    block = db.get_entry_credit_block(height=height)
    db.close()
    if block is None:
        bottle.abort(404)
    return block.to_dict()


@bottle.get(f"{RestPaths.ENTRY_CREDIT_BLOCK.value}/head")
def get_entry_credit_block_head():
    db = factom_core.db.FactomdLevelDB(create_if_missing=True)
    block = db.get_entry_credit_block_head()
    db.close()
    if block is None:
        bottle.abort(404)
    return block.to_dict()


@bottle.get(f"{RestPaths.ENTRY_BLOCK.value}/<keymr:re:{hex_regex}>")
def get_entry_block(keymr: str):
    db = factom_core.db.FactomdLevelDB(create_if_missing=True)
    block = db.get_entry_block(keymr=bytes.fromhex(keymr))
    db.close()
    if block is None:
        bottle.abort(404)
    return block.to_dict()


@bottle.get(f"{RestPaths.ENTRY_BLOCK.value}/<chain_id:re:{hex_regex}>/head")
def get_entry_block_head(chain_id: str):
    db = factom_core.db.FactomdLevelDB(create_if_missing=True)
    block = db.get_entry_block_head(chain_id=bytes.fromhex(chain_id))
    db.close()
    if block is None:
        bottle.abort(404)
    return block.to_dict()


@bottle.get(f"{RestPaths.ENTRY.value}/<entry_hash:re:{hex_regex}>")
def get_entry(entry_hash: str):
    db = factom_core.db.FactomdLevelDB(create_if_missing=True)
    entry = db.get_entry(bytes.fromhex(entry_hash))
    db.close()
    if entry is None:
        bottle.abort(404)
    return entry.to_dict()


@bottle.error(404)
def error404(e):
    body = {"errors": {"detail": "Object not found"}}
    return json.dumps(body, separators=(",", ":"))


def run():
    print("Starting API Server (localhost:8000)...")
    try:
        bottle.run(host="localhost", port=8000, quiet=True)
    except (KeyboardInterrupt, SystemExit):
        sys.exit()


if __name__ == "__main__":
    run()



from .admin_messages import AdminMessage
from .balance_increase import BalanceIncrease
from .chain import Chain
from .chain_commit import ChainCommit
from .entry import Entry
from .entry_commit import EntryCommit
from .factoid_transaction import FactoidTransaction
from .admin_messages import *  # TODO: stop being lazy, just explicitly import the classes



import struct
from dataclasses import dataclass

import factom_core.primitives as primitives
from factom_core.blockchains import Blockchain
from factom_core.blocks import DirectoryBlockHeader
from factom_core.messages import Message


@dataclass
class EndOfMinute(Message):
    """
    A message for leaders to send out signifying the end of the current block minute
    """

    TYPE = 0
    timestamp: bytes
    chain_id: bytes
    minute: int
    vm_index: int
    factoid_vm: int
    height: int
    system_height: int
    system_hash: bytes
    b: int
    signature: primitives.FullSignature

    def __post_init__(self):
        # TODO: type/value assertions
        self.is_p2p = True
        super().__post_init__()

    def marshal(self) -> bytes:
        """
        Marshal the message into the following representation:
        - first byte is the message type (always 10)
        - next 6 bytes are the timestamp
        - next 32 bytes are the server's identity chain id
        - next byte is the minute number
        - next byte is the vm index
        - next byte is the factoid vm index
        - next 4 bytes are the directory block height
        - next 4 bytes are the system height
        - next 32 bytes are the system hash
        - TODO: next byte is unknown.. but if its greater than 0, then the pubkey + signature must be next
        - next 32 bytes are the signing public key
        - next 64 bytes are the signature

        :return: byte representation of the message
        """
        buf = bytearray()
        buf.append(self.TYPE)
        buf.extend(self.timestamp)
        buf.extend(self.chain_id)
        buf.append(self.vm_index)
        buf.extend(struct.pack(">I", self.height))
        buf.extend(struct.pack(">I", self.system_height))
        buf.append(self.b)
        if self.b > 0:
            buf.extend(self.signature.marshal())
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        msg_type, data = raw[0], raw[1:]
        if msg_type != cls.TYPE:
            raise ValueError("Invalid message type ({})".format(msg_type))

        timestamp, data = data[:6], data[6:]
        chain_id, data = data[:32], data[32:]
        minute, data = data[0], data[1:]
        vm_index, data = data[0], data[1:]
        factoid_vm, data = data[0], data[1:]
        height, data = struct.unpack(">I", data[:4])[0], data[4:]
        system_height, data = struct.unpack(">I", data[:4])[0], data[4:]
        system_hash, data = data[:32], data[32:]
        b, data = data[0], data[1:]
        if b > 0:
            signature, data = primitives.FullSignature.unmarshal(data[:96]), data[96:]
        else:
            signature = primitives.FullSignature(bytes(32), bytes(64))

        return EndOfMinute(
            timestamp=timestamp,
            chain_id=chain_id,
            minute=minute,
            vm_index=vm_index,
            factoid_vm=factoid_vm,
            height=height,
            system_height=system_height,
            system_hash=system_hash,
            b=b,
            signature=signature,
        )

    def to_dict(self) -> dict:
        return {
            "timestamp": self.timestamp.hex(),
            "chain_id": self.chain_id.hex(),
            "minute": self.minute,
            "vm_index": self.vm_index,
            "factoid_vm": self.factoid_vm,
            "height": self.height,
            "system_height": self.system_height,
            "system_hash": self.system_hash.hex(),
            "b": self.b,
            "signature": self.signature.to_dict(),
        }

    def leader_execute(self, state: Blockchain):
        pass

    def follower_execute(self, state: Blockchain):
        pass


@dataclass
class DirectoryBlockSignature(Message):
    """
    A message for leaders to send out at the end of a block. Signs the header of the last built block.
    """

    TYPE = 7

    timestamp: bytes
    system_height: int
    system_hash: bytes
    height: int
    vm_index: int
    header: DirectoryBlockHeader
    chain_id: bytes
    header_signature: primitives.FullSignature  # over just dblock header, goes in the admin block
    signature: primitives.FullSignature  # over the message

    def __post_init__(self):
        # TODO: type/value assertions
        self.is_p2p = True
        super().__post_init__()

    def marshal(self) -> bytes:
        """
        Marshal the message into the following representation:
        - first byte is the message type (always 7)
        - next 6 bytes are the timestamp
        - next 4 bytes are the system height
        - next 32 bytes are the system hash
        - next 4 bytes are the directory block height
        - next byte is the vm index
        - next bytes are the marshalled directory block header
        - next 32 bytes are the server's identity chain id
        - next 32 bytes are the header signing public key (actual dbsig, goes in next adminblock)
        - next 64 bytes are the header signature (actual dbsig, goes in next adminblock)
        - next 32 bytes are the signing public key
        - next 64 bytes are the signature

        :return: byte representation of the message
        """
        buf = bytearray()
        buf.append(self.TYPE)
        buf.extend(self.timestamp)
        buf.extend(struct.pack(">I", self.system_height))
        buf.extend(self.system_hash)
        buf.extend(struct.pack(">I", self.height))
        buf.append(self.vm_index)
        buf.extend(self.header.marshal())
        buf.extend(self.chain_id)
        buf.extend(self.header_signature.marshal())
        buf.extend(self.signature.marshal())
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        msg_type, data = raw[0], raw[1:]
        if msg_type != cls.TYPE:
            raise ValueError("Invalid message type ({})".format(msg_type))

        timestamp, data = data[:6], data[6:]
        system_height, data = struct.unpack(">I", data[:4])[0], data[4:]
        system_hash, data = data[:32], data[32:]
        height, data = struct.unpack(">I", data[:4])[0], data[4:]
        vm_index, data = data[0], data[1:]
        header_data, data = (
            data[: DirectoryBlockHeader.LENGTH],
            data[DirectoryBlockHeader.LENGTH :],
        )
        header = DirectoryBlockHeader.unmarshal(header_data)
        chain_id, data = data[:32], data[32:]

        header_signature, data = (
            primitives.FullSignature.unmarshal(data[:96]),
            data[96:],
        )
        signature, data = primitives.FullSignature.unmarshal(data[:96]), data[96:]

        return DirectoryBlockSignature(
            timestamp=timestamp,
            system_height=system_height,
            system_hash=system_hash,
            height=height,
            vm_index=vm_index,
            header=header,
            chain_id=chain_id,
            header_signature=header_signature,
            signature=signature,
        )

    def to_dict(self) -> dict:
        return {
            "timestamp": self.timestamp.hex(),
            "system_height": self.system_height,
            "system_hash": self.system_hash.hex(),
            "height": self.height,
            "vm_index": self.vm_index,
            "header": self.header.__str__(),  # TODO: header to_dict() functions
            "chain_id": self.chain_id.hex(),
            "header_signature": self.header_signature.to_dict(),
            "signature": self.signature.to_dict(),
        }

    def leader_execute(self, state: Blockchain):
        pass

    def follower_execute(self, state: Blockchain):
        pass



import hashlib
import struct
from dataclasses import dataclass, field

import factom_core
from factom_core.utils import merkle


@dataclass
class DirectoryBlockHeader:
    LENGTH = 113

    CHAIN_ID = bytes.fromhex("000000000000000000000000000000000000000000000000000000000000000d")

    network_id: bytes
    body_mr: bytes
    prev_keymr: bytes
    prev_full_hash: bytes
    timestamp: int
    height: int
    block_count: int

    def __post_init__(self):
        # TODO: value assertions
        pass

    def marshal(self):
        buf = bytearray()
        buf.append(0x00)
        buf.extend(self.network_id)
        buf.extend(self.body_mr)
        buf.extend(self.prev_keymr)
        buf.extend(self.prev_full_hash)
        buf.extend(struct.pack(">I", int(self.timestamp / 60)))
        buf.extend(struct.pack(">I", self.height))
        buf.extend(struct.pack(">I", self.block_count))
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        if len(raw) != DirectoryBlockHeader.LENGTH:
            raise ValueError("`raw` must be exactly {} bytes long".format(DirectoryBlockHeader.LENGTH))
        data = raw[1:]  # skip single byte version
        network_id, data = data[:4], data[4:]
        body_mr, data = data[:32], data[32:]
        prev_keymr, data = data[:32], data[32:]
        prev_full_hash, data = data[:32], data[32:]
        timestamp, data = struct.unpack(">I", data[:4])[0], data[4:]
        timestamp *= 60  # timestamp in minutes, multiply by 60
        height, data = struct.unpack(">I", data[:4])[0], data[4:]
        block_count, data = struct.unpack(">I", data[:4])[0], data[4:]
        return DirectoryBlockHeader(
            network_id=network_id,
            body_mr=body_mr,
            prev_keymr=prev_keymr,
            prev_full_hash=prev_full_hash,
            timestamp=timestamp,
            height=height,
            block_count=block_count,
        )


@dataclass
class DirectoryBlockBody:

    admin_block_lookup_hash: bytes
    entry_credit_block_header_hash: bytes
    factoid_block_keymr: bytes
    entry_blocks: list = field(default_factory=list)

    _cached_mr: bytes = None

    def __post_init__(self):
        # TODO: value assertions
        pass

    @property
    def merkle_root(self):
        if self._cached_mr is not None:
            return self._cached_mr

        body_elements = [
            factom_core.blocks.AdminBlockHeader.CHAIN_ID,
            self.admin_block_lookup_hash,
            factom_core.blocks.EntryCreditBlockHeader.CHAIN_ID,
            self.entry_credit_block_header_hash,
            factom_core.blocks.FactoidBlockHeader.CHAIN_ID,
            self.factoid_block_keymr,
        ]
        for e_block in self.entry_blocks:
            body_elements.append(e_block.get("chain_id"))
            body_elements.append(e_block.get("keymr"))
        self._cached_mr = merkle.get_merkle_root(body_elements)
        return self._cached_mr

    def marshal(self):
        buf = bytearray()
        buf.extend(factom_core.blocks.AdminBlockHeader.CHAIN_ID)
        buf.extend(self.admin_block_lookup_hash)
        buf.extend(factom_core.blocks.EntryCreditBlockHeader.CHAIN_ID)
        buf.extend(self.entry_credit_block_header_hash)
        buf.extend(factom_core.blocks.FactoidBlockHeader.CHAIN_ID)
        buf.extend(self.factoid_block_keymr)
        for e_block in self.entry_blocks:
            buf.extend(e_block.get("chain_id"))
            buf.extend(e_block.get("keymr"))
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes, block_count: int):
        body, data = cls.unmarshal_with_remainder(raw, block_count)
        assert len(data) == 0, "Extra bytes remaining!"
        return body

    @classmethod
    def unmarshal_with_remainder(cls, raw: bytes, block_count: int):
        data = raw
        admin_block_chain_id, data = data[:32], data[32:]
        assert admin_block_chain_id == factom_core.blocks.AdminBlockHeader.CHAIN_ID
        admin_block_lookup_hash, data = data[:32], data[32:]
        ec_block_chain_id, data = data[:32], data[32:]
        assert ec_block_chain_id == factom_core.blocks.EntryCreditBlockHeader.CHAIN_ID
        entry_credit_block_header_hash, data = data[:32], data[32:]
        factoid_block_chain_id, data = data[:32], data[32:]
        assert factoid_block_chain_id == factom_core.blocks.FactoidBlockHeader.CHAIN_ID
        factoid_block_keymr, data = data[:32], data[32:]
        entry_blocks = []
        for i in range(block_count - 3):
            entry_block_chain_id, data = data[:32], data[32:]
            entry_block_keymr, data = data[:32], data[32:]
            entry_blocks.append({"chain_id": entry_block_chain_id, "keymr": entry_block_keymr})
        return (
            DirectoryBlockBody(
                admin_block_lookup_hash=admin_block_lookup_hash,
                entry_credit_block_header_hash=entry_credit_block_header_hash,
                factoid_block_keymr=factoid_block_keymr,
                entry_blocks=entry_blocks,
            ),
            data,
        )

    def construct_header(
        self, network_id: bytes, prev_keymr: bytes, prev_full_hash: bytes, timestamp: int, height: int,
    ) -> DirectoryBlockHeader:
        return DirectoryBlockHeader(
            network_id=network_id,
            body_mr=self.merkle_root,
            prev_keymr=prev_keymr,
            prev_full_hash=prev_full_hash,
            timestamp=timestamp,
            height=height,
            block_count=len(self.entry_blocks) + 3,
        )


@dataclass
class DirectoryBlock:

    header: DirectoryBlockHeader
    body: DirectoryBlockBody

    _cached_keymr: bytes = None

    def __post_init__(self,):
        # TODO: assert they're all here\
        pass

    @property
    def keymr(self):
        if self._cached_keymr is not None:
            return self._cached_keymr

        self._cached_keymr = merkle.calculate_keymr(self.header.marshal(), self.body.merkle_root)
        return self._cached_keymr

    @property
    def full_hash(self):
        return hashlib.sha256(self.marshal()).digest()

    def marshal(self):
        """Marshals the directory block according to the byte-level representation shown at
        https://github.com/FactomProject/FactomDocs/blob/master/factomDataStructureDetails.md#directory-block

        Data returned does not include contextual metadata, such as anchor information or the pointer to the
        next directory block.
        """
        buf = bytearray()
        buf.extend(self.header.marshal())
        buf.extend(self.body.marshal())
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        """Returns a new DirectoryBlock object, unmarshalling given bytes according to:
        https://github.com/FactomProject/FactomDocs/blob/master/factomDataStructureDetails.md#directory-block

        Useful for working with a single dblock out of context, pulled directly from a factomd database for instance.

        DirectoryBlock created will not include contextual metadata, such as anchor information or the pointer to the
        next directory block.
        """
        block, data = cls.unmarshal_with_remainder(raw)
        assert len(data) == 0, "Extra bytes remaining!"
        return block

    @classmethod
    def unmarshal_with_remainder(cls, raw: bytes):
        header_data, data = (
            raw[: DirectoryBlockHeader.LENGTH],
            raw[DirectoryBlockHeader.LENGTH :],
        )
        header = DirectoryBlockHeader.unmarshal(header_data)
        body, data = DirectoryBlockBody.unmarshal_with_remainder(data, header.block_count)
        return DirectoryBlock(header=header, body=body), data

    def to_dict(self):
        return {
            "keymr": self.keymr.hex(),
            "network_id": self.header.network_id.hex(),
            "body_mr": self.header.body_mr.hex(),
            "prev_keymr": self.header.prev_keymr.hex(),
            "prev_full_hash": self.header.prev_full_hash.hex(),
            "timestamp": self.header.timestamp,
            "height": self.header.height,
            "admin_block_lookup_hash": self.body.admin_block_lookup_hash.hex(),
            "entry_credit_block_header_hash": self.body.entry_credit_block_header_hash.hex(),
            "factoid_block_keymr": self.body.factoid_block_keymr.hex(),
            "entry_blocks": [
                {"chain_id": entry_block.get("chain_id").hex(), "keymr": entry_block.get("keymr").hex(),}
                for entry_block in self.body.entry_blocks
            ],
        }

    def __str__(self):
        return "{}(height={}, keymr={})".format(self.__class__.__name__, self.header.height, self.keymr.hex())



from .admin_block import AdminBlock, AdminBlockBody, AdminBlockHeader
from .directory_block import DirectoryBlock, DirectoryBlockBody, DirectoryBlockHeader
from .entry_block import EntryBlock, EntryBlockBody, EntryBlockHeader
from .entry_credit_block import (
    EntryCreditBlock,
    EntryCreditBlockBody,
    EntryCreditBlockHeader,
)
from .factoid_block import FactoidBlock, FactoidBlockBody, FactoidBlockHeader



from dataclasses import dataclass

import factom_core.block_elements as block_elements
import factom_core.primitives as primitives
from factom_core.blockchains import Blockchain
from factom_core.messages import Message


@dataclass
class FactoidTransaction(Message):
    """
    A message holding a FactoidTransaction object.
    """

    TYPE = 9

    tx: block_elements.FactoidTransaction

    def __post_init__(self):
        # TODO: type/value assertions
        super().__post_init__()

    def marshal(self) -> bytes:
        """
        Marshal the message into the following representation:
        - first byte is the message type (always 9)
        - next bytes are the marshalled Factoid Transaction object

        :return: byte representation of the message
        """
        buf = bytearray()
        buf.append(self.TYPE)
        buf.extend(self.tx.marshal())
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        msg_type, data = raw[0], raw[1:]
        if msg_type != cls.TYPE:
            raise ValueError("Invalid message type ({})".format(msg_type))
        tx = block_elements.FactoidTransaction.unmarshal(data)
        return FactoidTransaction(tx=tx)

    def to_dict(self) -> dict:
        return {"tx": self.tx.to_dict()}

    def leader_execute(self, state: Blockchain):
        pass

    def follower_execute(self, state: Blockchain):
        pass


@dataclass
class ChainCommit(Message):
    """
    A message holding a ChainCommit object.
    """

    TYPE = 5

    commit: block_elements.ChainCommit
    signature: primitives.FullSignature

    def __post_init__(self):
        # TODO: type/value assertions
        super().__post_init__()

    def marshal(self) -> bytes:
        """
        Marshal the message into the following representation:
        - first byte is the message type (always 5)
        - next bytes are the marshalled Chain Commit object
        - next 32 bytes are the public key
        - next 64 bytes are the signature

        :return: byte representation of the message
        """
        buf = bytearray()
        buf.append(self.TYPE)
        buf.extend(self.commit.marshal())
        buf.extend(self.signature.marshal())
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        msg_type, data = raw[0], raw[1:]
        if msg_type != cls.TYPE:
            raise ValueError("Invalid message type ({})".format(msg_type))

        commit_size = block_elements.ChainCommit.BITLENGTH
        commit, data = (
            block_elements.ChainCommit.unmarshal(data[:commit_size]),
            data[commit_size:],
        )

        # TODO: found in factomd code, is this actually not signed sometimes?
        if len(data) > 0:
            signature, data = primitives.FullSignature.unmarshal(data[:96]), data[96:]
        else:
            signature = primitives.FullSignature(b"", b"")

        return ChainCommit(commit=commit, signature=signature)

    def to_dict(self) -> dict:
        return {"commit": self.commit.to_dict(), "signature": self.signature.to_dict()}

    def leader_execute(self, state: Blockchain):
        pass

    def follower_execute(self, state: Blockchain):
        pass


@dataclass
class EntryCommit(Message):
    """
    A message holding a EntryCommit object.
    """

    TYPE = 6

    commit: block_elements.EntryCommit
    signature: primitives.FullSignature

    def __post_init__(self):
        # TODO: type/value assertions
        super().__post_init__()

    def marshal(self) -> bytes:
        """
        Marshal the message into the following representation:
        - first byte is the message type (always 6)
        - next bytes are the marshalled Entry Commit object
        - next 32 bytes are the public key
        - next 64 bytes are the signature

        :return: byte representation of the message
        """
        buf = bytearray()
        buf.append(self.TYPE)
        buf.extend(self.commit.marshal())
        buf.extend(self.signature.marshal())
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        msg_type, data = raw[0], raw[1:]
        if msg_type != cls.TYPE:
            raise ValueError("Invalid message type ({})".format(msg_type))

        commit_size = block_elements.EntryCommit.BITLENGTH
        commit, data = (
            block_elements.EntryCommit.unmarshal(data[:commit_size]),
            data[commit_size:],
        )

        if len(data) > 0:  # TODO: found in factomd code, is this actually not signed sometimes?
            signature, data = primitives.FullSignature.unmarshal(data[:96]), data[96:]
        else:
            signature = primitives.FullSignature(b"", b"")

        return EntryCommit(commit=commit, signature=signature)

    def to_dict(self) -> dict:
        return {"commit": self.commit.to_dict(), "signature": self.signature.to_dict()}

    def leader_execute(self, state: Blockchain):
        pass

    def follower_execute(self, state: Blockchain):
        pass


@dataclass
class EntryReveal(Message):
    """
    A message holding an Entry object to be revealed (can reveal a Chain as well)
    """

    TYPE = 13

    timestamp: bytes
    entry: block_elements.Entry

    def __post_init__(self):
        # TODO: type/value assertions
        super().__post_init__()

    def marshal(self) -> bytes:
        """
        Marshal the message into the following representation:
        - first byte is the message type (always 13)
        - next 6 bytes are the timestamp
        - next bytes are the marshalled Entry itself

        :return: byte representation of the message
        """
        buf = bytearray()
        buf.append(self.TYPE)
        buf.extend(self.timestamp)
        buf.extend(self.entry.marshal())
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        msg_type, data = raw[0], raw[1:]
        if msg_type != cls.TYPE:
            raise ValueError("Invalid message type ({})".format(msg_type))

        timestamp, data = data[:6], data[6:]
        entry = block_elements.Entry.unmarshal(data)
        return EntryReveal(timestamp=timestamp, entry=entry)

    def to_dict(self) -> dict:
        return {"timestamp": self.timestamp.hex(), "entry": self.entry.to_dict()}

    def leader_execute(self, state: Blockchain):
        pass

    def follower_execute(self, state: Blockchain):
        pass



import struct
import hashlib
from dataclasses import dataclass
from typing import Dict, List

from factom_core.utils import merkle
from .directory_block import DirectoryBlock


@dataclass
class EntryBlockHeader:
    LENGTH = 140

    """Header section of an Entry Block, built from the body or unmarshalled from raw bytes"""

    chain_id: bytes
    body_mr: bytes
    prev_keymr: bytes
    prev_full_hash: bytes
    sequence: int
    height: int
    entry_count: int

    def marshal(self) -> bytes:
        buf = bytearray()
        buf.extend(self.chain_id)
        buf.extend(self.body_mr)
        buf.extend(self.prev_keymr)
        buf.extend(self.prev_full_hash)
        buf.extend(struct.pack(">I", self.sequence))
        buf.extend(struct.pack(">I", self.height))
        buf.extend(struct.pack(">I", self.entry_count))
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        if len(raw) != EntryBlockHeader.LENGTH:
            raise ValueError("`raw` must be exactly {} bytes long".format(EntryBlockHeader.LENGTH))
        chain_id, data = raw[:32], raw[32:]
        body_mr, data = data[:32], data[32:]
        prev_keymr, data = data[:32], data[32:]
        prev_full_hash, data = data[:32], data[32:]
        sequence, data = struct.unpack(">I", data[:4])[0], data[4:]
        height, data = struct.unpack(">I", data[:4])[0], data[4:]
        entry_count, data = struct.unpack(">I", data[:4])[0], data[4:]
        return EntryBlockHeader(
            chain_id=chain_id,
            body_mr=body_mr,
            prev_keymr=prev_keymr,
            prev_full_hash=prev_full_hash,
            sequence=sequence,
            height=height,
            entry_count=entry_count,
        )


@dataclass
class EntryBlockBody:

    entry_hashes: Dict[int, List[bytes]]
    _cached_mr: bytes = None

    def __post_init__(self):
        # TODO: assert they're all here
        pass

    @property
    def merkle_root(self):
        if self._cached_mr is not None:
            return self._cached_mr

        body_elements = []
        for minute, hashes in self.entry_hashes.items():
            for h in hashes:
                body_elements.append(h)
            buf = bytearray()
            buf.extend(bytes(31))
            buf.append(minute)
            body_elements.append(bytes(buf))
        self._cached_mr = merkle.get_merkle_root(body_elements)
        return self._cached_mr

    def marshal(self):
        buf = bytearray()
        for minute, hashes in self.entry_hashes.items():
            for h in hashes:
                buf.extend(h)
            buf.extend(bytes(31))
            buf.append(minute)
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes, entry_count: int):
        body, data = cls.unmarshal_with_remainder(raw, entry_count)
        assert len(data) == 0, "Extra bytes remaining!"
        return body

    @classmethod
    def unmarshal_with_remainder(cls, raw: bytes, entry_count: int):
        data = raw
        # Entry hashes are listed in order, with a minute marker following what minute those entries were in
        entry_hashes = {}
        current_minute_entries = []
        for i in range(entry_count):
            entry_hash, data = data[:32], data[32:]
            if entry_hash[:-1] == bytes(31) and entry_hash[-1] <= 10:
                entry_hashes[entry_hash[-1]] = current_minute_entries
                current_minute_entries = []
            else:
                current_minute_entries.append(entry_hash)

        return EntryBlockBody(entry_hashes=entry_hashes), data

    def construct_header(
        self, chain_id: bytes, prev_keymr: bytes, prev_full_hash: bytes, sequence: int, height: int,
    ) -> EntryBlockHeader:
        """
        Creates an returns an EntryBlockHeader for this body object, given the specified contextual
        parameters.
        """
        entry_count = 0
        for hashes in self.entry_hashes.values():
            entry_count += len(hashes)
        return EntryBlockHeader(
            chain_id=chain_id,
            body_mr=self.merkle_root,
            prev_keymr=prev_keymr,
            prev_full_hash=prev_full_hash,
            sequence=sequence,
            height=height,
            entry_count=entry_count,
        )


@dataclass
class EntryBlock:

    header: EntryBlockHeader
    body: EntryBlockBody

    _cached_keymr: bytes = None

    directory_block_keymr: bytes = None
    timestamp: int = None

    def __post_init__(self):
        # TODO: assert they're all here
        pass
        # Optional contextual metadata. Derived from the directory block that contains this EntryBlock
        # self.next_keymr = kwargs.get("next_keymr")

    @property
    def keymr(self):
        if self._cached_keymr is not None:
            return self._cached_keymr

        self._cached_keymr = merkle.calculate_keymr(self.header.marshal(), self.body.merkle_root)
        return self._cached_keymr

    @property
    def full_hash(self):
        return hashlib.sha256(self.marshal()).digest()

    def marshal(self):
        """Marshals the entry block according to the byte-level representation shown at
        https://github.com/FactomProject/FactomDocs/blob/master/factomDataStructureDetails.md#entry-block

        Data returned does not include contextual metadata, such as timestamp or the pointer to the next entry block.
        """
        buf = bytearray()
        buf.extend(self.header.marshal())
        buf.extend(self.body.marshal())
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        """Returns a new EntryBlock object, unmarshalling given bytes according to:
        https://github.com/FactomProject/FactomDocs/blob/master/factomDataStructureDetails.md#entry-block

        Useful for working with a single eblock out of context, pulled directly from a factomd database for instance.

        EntryBlock created will not include contextual metadata, such as timestamp or the pointer to the
        next entry block.
        """
        block, data = cls.unmarshal_with_remainder(raw)
        assert len(data) == 0, "Extra bytes remaining!"
        return block

    @classmethod
    def unmarshal_with_remainder(cls, raw: bytes):
        header_data, data = (
            raw[: EntryBlockHeader.LENGTH],
            raw[EntryBlockHeader.LENGTH :],
        )
        header = EntryBlockHeader.unmarshal(header_data)
        body, data = EntryBlockBody.unmarshal_with_remainder(data, header.entry_count)
        return EntryBlock(header=header, body=body), data

    def add_context(self, directory_block: DirectoryBlock):
        self.directory_block_keymr = directory_block.keymr
        self.timestamp = directory_block.header.timestamp

    def to_dict(self):
        return {
            # Required
            "keymr": self.keymr.hex(),
            "chain_id": self.header.chain_id.hex(),
            "prev_keymr": self.header.prev_keymr.hex(),
            "prev_full_hash": self.header.prev_full_hash.hex(),
            "sequence": self.header.sequence,
            "height": self.header.height,
            "entry_hashes": {minute: [h.hex() for h in hashes] for minute, hashes in self.body.entry_hashes.items()},
            # Optional contextual
            "directory_block_keymr": None if self.directory_block_keymr is None else self.directory_block_keymr.hex(),
            "timestamp": self.timestamp,
        }

    def __str__(self):
        return "{}(height={}, keymr={})".format(self.__class__.__name__, self.header.height, self.keymr.hex())



import struct
from dataclasses import dataclass
from factom_core.blocks.entry_block import EntryBlock
from hashlib import sha256, sha512


@dataclass
class Entry:

    chain_id: bytes
    external_ids: list
    content: bytes

    _cached_entry_hash: bytes = None

    directory_block_keymr: bytes = None
    entry_block_keymr: bytes = None
    height: int = None
    timestamp: int = None
    stage: str = None

    def __post_init__(self):
        # TODO: value assertions
        pass
        # Optional contextual metadata. Derived from the directory block that contains this EntryBlock

    @property
    def entry_hash(self):
        """The entry hash in bytes. The algorithm used, along with the rationale behind its use, is shown at:
        https://github.com/FactomProject/FactomDocs/blob/master/factomDataStructureDetails.md#entry-hash
        """
        if self._cached_entry_hash is not None:
            return self._cached_entry_hash

        # Entry Hash = SHA256(SHA512(marshalled_entry_data) + marshalled_entry_data)
        data = self.marshal()
        h = sha512(data).digest()
        self._cached_entry_hash = sha256(h + data).digest()
        return self._cached_entry_hash

    def marshal(self):
        """Marshals the entry according to the byte-level representation shown at
        https://github.com/FactomProject/FactomDocs/blob/master/factomDataStructureDetails.md#entry

        Useful for proofs and constructing entry hashes.

        Data returned does not include contextual metadata, such as created_at, entry_block, directory_block, stage,
        and other information inferred from where the entry lies in its chain.
        """
        buf = bytearray()
        buf.append(0x00)  # single byte version
        buf.extend(self.chain_id)
        external_ids_size = 0
        ext_id_data = b""
        for external_id in self.external_ids:
            size = len(external_id)
            external_ids_size += size + 2
            ext_id_data += struct.pack(">h", size)
            ext_id_data += external_id
        size = struct.pack(">h", external_ids_size)
        buf.extend(size)
        buf.extend(ext_id_data)
        buf.extend(self.content)
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        """Returns a new Entry object, unmarshalling given bytes according to:
        https://github.com/FactomProject/FactomDocs/blob/master/factomDataStructureDetails.md#entry

        Useful for working with a single entry out of context, pulled directly from a factomd database for instance.

        Entry created will not include contextual metadata, such as created_at, entry_block, directory_block, stage, and
        other information inferred from where the entry lies in its chain.
        """
        data = raw[1:]  # skip single byte version, probably just gonna be 0x00 for a long time anyways
        chain_id, data = data[:32], data[32:]
        external_ids_size, data = struct.unpack(">h", data[:2])[0], data[2:]
        external_ids = []
        while external_ids_size > 0:
            size, data = struct.unpack(">h", data[:2])[0], data[2:]
            external_id, data = data[:size], data[size:]
            external_ids.append(external_id)
            external_ids_size = external_ids_size - size - 2
        content = data  # Leftovers are the entry content
        return Entry(chain_id=chain_id, external_ids=external_ids, content=content)

    def add_context(self, entry_block: EntryBlock):
        self.directory_block_keymr = entry_block.directory_block_keymr
        self.entry_block_keymr = entry_block.keymr
        self.height = entry_block.header.height
        # Find what minute this entry appeared in within the entry block
        base_timestamp = entry_block.timestamp
        for minute, entry_hashes in entry_block.entry_hashes.items():
            if self.entry_hash in entry_hashes:
                self.timestamp = base_timestamp + minute * 60
                break
        else:
            # Entry not found, raise an error
            raise ValueError("provided EntryBlock does not contain this entry")

    def to_dict(self):
        return {
            # Required
            "chain_id": self.chain_id.hex(),
            "entry_hash": self.entry_hash.hex(),
            "external_ids": [e.hex() for e in self.external_ids],
            "content": self.content.hex(),
            # Optional contextual
            "directory_block_keymr": None if self.directory_block_keymr is None else self.directory_block_keymr.hex(),
            "entry_block_keymr": None if self.entry_block_keymr is None else self.entry_block_keymr.hex(),
            "height": self.height,
            "timestamp": self.timestamp,
            "stage": self.stage,
        }

    def __str__(self):
        return "{}(chain_id={}, entry_hash={})".format(
            self.__class__.__name__, self.chain_id.hex(), self.entry_hash.hex()
        )



from dataclasses import dataclass, field

import factom_core.messages as messages
from factom_core.blockchains import Blockchain


@dataclass
class Message:
    """
    A message for VM execution.
    """

    TYPE = None  # type: int
    is_local: bool = field(init=False, default=False)

    # __slots__ = [
    #     "origin",  # set and examined on each server (not marshalled)
    #     "network_origin",  # hash of the network peer/connection where the message is from
    #     "is_p2p",  # (not marshalled)
    #     "is_local_only",  # if true, don't broadcast. we can skip verification too since it was locally sourced
    #     "is_from_network",  # if true, we got this message from the network (not marshalled)
    #     "is_full_broadcast",  # used for messages with no missing message support e.g. election related messages
    #     "is_resendable",
    #     "resend_limit",
    #     "leader_chain_id",
    #     "message_hash",  # cache of the hash of a message TODO: needed?
    #     "repeat_hash",  # cache of the hash of a message TODO: needed?
    #     "vm_index",  # index of the VM responsible for this message
    #     "vm_hash",  # basis for selecting the vm index
    #     "minute",
    #     "time_to_resend",  # milliseconds
    #     "time_to_expire",  # milliseconds
    #     "is_stalled",
    #     "is_invalid",
    #     "is_sig_valid",
    # ]

    def __post_init__(self):
        if self.TYPE is None:
            raise ValueError("A Message class must be instantiated with a `TYPE`")

    def leader_execute(self, state: Blockchain):
        pass

    def follower_execute(self, state: Blockchain):
        pass

    def marshal(self) -> bytes:
        pass

    @classmethod
    def unmarshal(cls, raw: bytes):
        pass

    def to_dict(self) -> dict:
        return {}


def unmarshal_message(raw: bytes):
    msg_type = raw[0]
    if msg_type == messages.Ack.TYPE:
        return messages.Ack.unmarshal(raw)

    elif msg_type == messages.ChainCommit.TYPE:
        return messages.ChainCommit.unmarshal(raw)

    elif msg_type == messages.EntryCommit.TYPE:
        return messages.EntryCommit.unmarshal(raw)

    elif msg_type == messages.EntryReveal.TYPE:
        return messages.EntryReveal.unmarshal(raw)

    elif msg_type == messages.FactoidTransaction.TYPE:
        return messages.FactoidTransaction.unmarshal(raw)

    elif msg_type == messages.EndOfMinute.TYPE:
        return messages.EndOfMinute.unmarshal(raw)

    elif msg_type == messages.DirectoryBlockSignature.TYPE:
        return messages.DirectoryBlockSignature.unmarshal(raw)

    elif msg_type == messages.DirectoryBlockStateRequest.TYPE:
        return messages.DirectoryBlockStateRequest.unmarshal(raw)

    elif msg_type == messages.DirectoryBlockState.TYPE:
        return messages.DirectoryBlockState.unmarshal(raw)

    elif msg_type == messages.MissingDataRequest.TYPE:
        return messages.MissingDataRequest.unmarshal(raw)

    elif msg_type == messages.MissingDataResponse.TYPE:
        return messages.MissingDataResponse.unmarshal(raw)

    elif msg_type == messages.Heartbeat.TYPE:
        return messages.Heartbeat.unmarshal(raw)

    elif msg_type == messages.AddServer.TYPE:
        return messages.AddServer.unmarshal(raw)

    elif msg_type == messages.ChangeServerKey.TYPE:
        return messages.ChangeServerKey.unmarshal(raw)

    elif msg_type == messages.RemoveServer.TYPE:
        return messages.RemoveServer.unmarshal(raw)

    elif msg_type == messages.BlockRequest.TYPE:
        return messages.BlockRequest.unmarshal(raw)

    elif msg_type == messages.MissingMessageRequest.TYPE:
        return messages.MissingMessageRequest.unmarshal(raw)

    elif msg_type == messages.MissingMessageResponse.TYPE:
        return messages.MissingMessageResponse.unmarshal(raw)

    else:
        raise ValueError("Bad message type")



from hashlib import sha256


def get_merkle_root(hashes: list) -> bytes:
    merkle_tree = build_merkle_tree(hashes)
    return None if len(merkle_tree) == 0 else merkle_tree[-1]


def build_merkle_tree(hashes: list):
    if len(hashes) == 0 or len(hashes) == 1:
        return hashes

    next_level = []
    for i in range(0, len(hashes), 2):
        left = hashes[i]
        right = hashes[i + 1] if i + 1 != len(hashes) else left
        top = sha256(left + right).digest()
        next_level.append(top)

    next_iteration = build_merkle_tree(next_level)
    hashes.extend(next_iteration)
    return hashes


def calculate_keymr(header: bytes, body_mr: bytes):
    header_hash = sha256(header).digest()
    return sha256(header_hash + body_mr).digest()



from setuptools import (
    setup,
    find_packages,
)


deps = {"factom-core": ["factom-keys", "plyvel",], "hydra": ["bottle", "click", "plyvel", "requests",]}

setup(
    name="factom-core",
    version="0.0.6",
    description="A python library for working with the primitives of the Factom blockchain",
    author="Sam Barnes",
    author_email="mistersamuelbarnes@gmail.com",
    url="https://github.com/sambarnes/factom-core",
    keywords=["factom", "core", "blockchain"],
    license="MIT",
    py_modules=["factom_core"],
    install_requires=deps["factom-core"],
    zip_safe=False,
    packages=find_packages(exclude=["tests", "tests.*", "hydra", "p2p"]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries",
        "Topic :: Utilities",
    ],
)



from dataclasses import dataclass


@dataclass
class EntryCommit:

    ECID = 0x03
    BITLENGTH = 136

    timestamp: int
    entry_hash: bytes
    ec_spent: int
    ec_public_key: bytes
    signature: bytes = None

    def __post_init__(self):
        # TODO: assert they're all here
        pass

    def marshal(self):
        """Marshals the EntryCommit according to the byte-level representation shown at
        https://github.com/FactomProject/FactomDocs/blob/master/factomDataStructureDetails.md#entry-commit
        """
        # Fail if signature is not set.
        assert self.signature is not None

        buf = bytearray()
        buf.extend(self.marshal_for_signature())
        buf.extend(self.ec_public_key)
        buf.extend(self.signature)
        return bytes(buf)

    def marshal_for_signature(self):
        buf = bytearray()
        buf.append(0x00)
        buf.extend(self.timestamp.to_bytes(6, "big", signed=False))
        buf.extend(self.entry_hash)
        buf.append(self.ec_spent)
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        """Returns a new EntryCommit object, unmarshalling given bytes according to:
        https://github.com/FactomProject/FactomDocs/blob/master/factomDataStructureDetails.md#entry-commit
        """
        data = raw[1:]  # skip single byte version, probably just 0x00 anyways
        timestamp, data = int.from_bytes(data[:6], "big", signed=False), data[6:]
        entry_hash, data = data[:32], data[32:]
        ec_spent, data = data[:1], data[1:]
        ec_spent = int.from_bytes(ec_spent, byteorder="big")
        assert ec_spent < 11, "Invalid EC spent ({}) for entry: {}".format(ec_spent, entry_hash.hex())
        ec_public_key, data = data[:32], data[32:]
        signature, data = data[:64], data[64:]  # covers version through ec spent
        assert len(data) == 0, "Extra bytes remaining!"

        return EntryCommit(
            timestamp=timestamp,
            entry_hash=entry_hash,
            ec_spent=ec_spent,
            ec_public_key=ec_public_key,
            signature=signature,
        )

    def to_dict(self):
        return {
            "timestamp": self.timestamp.hex(),
            "entry_hash": self.entry_hash.hex(),
            "ec_spent": self.ec_spent,
            "ec_public_key": self.ec_public_key.hex(),
            "signature": self.signature.hex(),
        }

    def __str__(self):
        # TODO: convert timestamp to readable and EC Public Key to its base58 address
        return "{}(timestamp={}, entry_hash={}, ec_public_key={})".format(
            self.__class__.__name__, self.timestamp, self.entry_hash.hex(), self.ec_public_key.hex(),
        )



from dataclasses import dataclass

from factom_core.block_elements import Entry
from factom_core.blockchains import Blockchain
from factom_core.blocks import EntryBlock
from factom_core.messages import Message


@dataclass
class MissingDataRequest(Message):
    """
    A request for an entry or entry block. Used when entry syncing.
    """

    TYPE = 17

    timestamp: bytes
    request_hash: bytes

    def __post_init__(self):
        # TODO: type/value assertions
        self.is_p2p = True
        super().__post_init__()

    def marshal(self) -> bytes:
        """
        Marshal the message into the following representation:
        - first byte is the message type (always 17)
        - next 6 bytes are the timestamp
        - next 32 bytes are the entry hash or entry block keymr being requested

        :return: byte representation of the message
        """
        buf = bytearray()
        buf.append(self.TYPE)
        buf.extend(self.timestamp)
        buf.extend(self.request_hash)
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        msg_type, data = raw[0], raw[1:]
        if msg_type != cls.TYPE:
            raise ValueError("Invalid message type ({})".format(msg_type))

        timestamp, data = data[:6], data[6:]
        request_hash, data = data[:32], data[32:]

        return MissingDataRequest(timestamp=timestamp, request_hash=request_hash)

    def to_dict(self):
        return {"timestamp": self.timestamp.hex(), "request_hash": self.request_hash}

    def __str__(self):
        return "{}(hash={})".format(self.__class__.__name__, self.request_hash)

    def leader_execute(self, state: Blockchain):
        self.follower_execute(state)

    def follower_execute(self, state: Blockchain):
        pass


@dataclass
class MissingDataResponse(Message):
    """
    A response to a MissingDataRequest, including the Entry or Entry Block requested.
    """

    TYPE = 18

    requested_object: object

    def __post_init__(self):
        if type(self.requested_object) not in {Entry, EntryBlock}:
            raise ValueError("requested_object must be an Entry or Entry Block")
        self.is_p2p = True
        super().__post_init__()

    def marshal(self) -> bytes:
        """
        Marshal the message into the following representation:
        - first byte is the message type (always 18)
        - second byte is the object type: 0 for Entry, 1 for Entry Block
        - remaining bytes are the marshalled Entry or Entry Block that was requested

        :return: byte representation of the message
        """
        buf = bytearray()
        buf.append(self.TYPE)
        buf.append(0 if isinstance(self.requested_object, Entry) else 1)
        buf.extend(self.requested_object.marshal())
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        msg_type, data = raw[0], raw[1:]
        if msg_type != cls.TYPE:
            raise ValueError("Invalid message type ({})".format(msg_type))

        object_type, data = data[0], data[1:]
        if object_type == 0:  # Entry
            requested_object = Entry.unmarshal(data)
        elif object_type == 1:  # Entry Block
            requested_object = EntryBlock.unmarshal(data)
        else:
            raise ValueError("Invalid object type ({})".format(object_type))

        return MissingDataResponse(requested_object=requested_object)

    def to_dict(self):
        assert type(self.requested_object) in {Entry, EntryBlock}
        return {"requested_object": self.requested_object.to_dict()}

    def __str__(self):
        h = (
            self.requested_object.entry_hash
            if isinstance(self.requested_object, Entry)
            else self.requested_object.keymr
        )
        return f"{self.__class__.__name__}(type={self.requested_object.__class__.__name__}, hash={h})"

    def leader_execute(self, state: Blockchain):
        self.follower_execute(state)

    def follower_execute(self, state: Blockchain):
        pass



from dataclasses import dataclass

import factom_core.primitives as primitives
from factom_core.blockchains import Blockchain
from factom_core.messages import Message


@dataclass
class AddServer(Message):
    """
    Add the identity as a Federated or Audit server
    """

    TYPE = 22

    timestamp: bytes
    chain_id: bytes
    is_federated: bool
    signature: primitives.FullSignature

    def __post_init__(self):
        # TODO: type/value assertions
        self.is_p2p = True
        super().__post_init__()

    def marshal(self) -> bytes:
        """
        Marshal the message into the following representation:
        - first byte is the message type (always 22)
        - next 6 bytes are a timestamp
        - next 32 bytes are the server's identity chain id
        - next byte is the server type (0 == Federated, 1 == Audit)
        - next 32 bytes are the public key signing this message
        - next 64 bytes are the signature

        :return: byte representation of the message
        """
        buf = bytearray()
        buf.append(self.TYPE)
        buf.extend(self.timestamp)
        buf.extend(self.chain_id)
        buf.append(0 if self.is_federated else 1)
        buf.extend(self.signature.marshal())
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        msg_type, data = raw[0], raw[1:]
        if msg_type != cls.TYPE:
            raise ValueError("Invalid message type ({})".format(msg_type))

        timestamp, data = data[:6], data[6:]
        chain_id, data = data[:32], data[32:]
        server_type, data = data[0], data[1:]
        signature, data = primitives.FullSignature.unmarshal(data[:96]), data[96:]
        assert len(data) == 0, "Extra bytes remaining!"
        return AddServer(
            timestamp=timestamp,
            chain_id=chain_id,
            is_federated=(server_type == 0),  # 0 == Federated, 1 == Audit
            signature=signature,
        )

    def to_dict(self) -> dict:
        return {
            "timestamp": self.timestamp.hex(),
            "chain_id": self.chain_id.hex(),
            "is_federated": self.is_federated,
            "signature": self.signature.to_dict(),
        }

    def leader_execute(self, state: Blockchain):
        pass

    def follower_execute(self, state: Blockchain):
        pass


@dataclass
class ChangeServerKey(Message):
    """
    Change the key for the specified server identity
    """

    TYPE = 23

    CHANGE_TYPE_ADD_MATRYOSHKA = 0x03
    CHANGE_TYPE_ADD_FED_SERVER_KEY = 0x08
    CHANGE_TYPE_ADD_BTC_ANCHOR_KEY = 0x09

    timestamp: bytes
    chain_id: bytes
    admin_block_change: int
    key_type: int
    priority: int
    new_key: bytes
    signature: primitives.FullSignature

    def __post_init__(self):
        # TODO: type/value assertions
        super().__post_init__()

    def marshal(self) -> bytes:
        """
        Marshal the message into the following representation:
        - first byte is the message type (always 23)
        - next 6 bytes are a timestamp
        - next 32 bytes are the server's identity chain id
        - next byte is Admin Block Change
        - next byte is the key type
        - next byte is the key priority
        - next 32 bytes are the new public key for the identity
        - next 32 bytes are the public key signing this message
        - next 64 bytes are the signature

        :return: byte representation of the message
        """
        buf = bytearray()
        buf.append(self.TYPE)
        buf.extend(self.timestamp)
        buf.extend(self.chain_id)
        buf.append(self.admin_block_change)
        buf.append(self.key_type)
        buf.append(self.priority)
        buf.extend(self.new_key)
        buf.extend(self.signature.marshal())
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        msg_type, data = raw[0], raw[1:]
        if msg_type != cls.TYPE:
            raise ValueError("Invalid message type ({})".format(msg_type))

        timestamp, data = data[:6], data[6:]
        chain_id, data = data[:32], data[32:]
        admin_block_change, data = data[0], data[1:]
        key_type, data = data[0], data[1:]
        priority, data = data[0], data[1:]
        new_key, data = data[:32], data[32:]
        signature, data = primitives.FullSignature.unmarshal(data[:96]), data[96:]
        assert len(data) == 0, "Extra bytes remaining!"
        return ChangeServerKey(
            timestamp=timestamp,
            chain_id=chain_id,
            admin_block_change=admin_block_change,
            key_type=key_type,
            priority=priority,
            new_key=new_key,
            signature=signature,
        )

    def to_dict(self) -> dict:
        return {
            "timestamp": self.timestamp.hex(),
            "chain_id": self.chain_id.hex(),
            "admin_block_change": self.admin_block_change,
            "key_type": self.key_type,
            "priority": self.priority,
            "new_key": self.new_key.hex(),
            "signature": self.signature.to_dict(),
        }

    def leader_execute(self, state: Blockchain):
        pass

    def follower_execute(self, state: Blockchain):
        pass


@dataclass
class RemoveServer(Message):
    """
    Remove the identity as a Federated or Audit server
    """

    TYPE = 24

    timestamp: bytes
    chain_id: bytes
    is_federated: bool
    signature: primitives.FullSignature

    def __post_init__(self):
        # TODO: type/value assertions
        self.is_p2p = True
        super().__post_init__()

    def marshal(self) -> bytes:
        """
        Marshal the message into the following representation:
        - first byte is the message type (always 24)
        - next 6 bytes are a timestamp
        - next 32 bytes are the server's identity chain id
        - next byte is the server type (0 == Federated, 1 == Audit)
        - next 32 bytes are the public key signing this message
        - next 64 bytes are the signature

        :return: byte representation of the message
        """
        buf = bytearray()
        buf.append(self.TYPE)
        buf.extend(self.timestamp)
        buf.extend(self.chain_id)
        buf.append(0 if self.is_federated else 1)
        buf.extend(self.signature.marshal())
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        msg_type, data = raw[0], raw[1:]
        if msg_type != cls.TYPE:
            raise ValueError("Invalid message type ({})".format(msg_type))

        timestamp, data = data[:6], data[6:]
        chain_id, data = data[:32], data[32:]
        server_type, data = data[0], data[1:]
        signature, data = primitives.FullSignature.unmarshal(data[:96]), data[96:]
        assert len(data) == 0, "Extra bytes remaining!"
        return RemoveServer(
            timestamp=timestamp,
            chain_id=chain_id,
            is_federated=(server_type == 0),  # 0 == Federated, 1 == Audit
            signature=signature,
        )

    def to_dict(self) -> dict:
        return {
            "timestamp": self.timestamp.hex(),
            "chain_id": self.chain_id.hex(),
            "is_federated": self.is_federated,
            "signature": self.signature.to_dict(),
        }

    def leader_execute(self, state: Blockchain):
        pass

    def follower_execute(self, state: Blockchain):
        pass
